# # https://www.codetd.com/en/article/7238534
# import re
# from sys import exec_prefix
# from xml.dom import NotFoundErr
# from xml.etree.ElementInclude import include
# import win32gui, win32con, win32com.client
# from time import sleep, time


# class cWindow:
#     def __init__(self):
#         self._hwnd = None
#         self.shell = win32com.client.Dispatch("WScript.Shell")
#         self._folder_control_id = None
#     def BringToTop(self):
#         win32gui.BringWindowToTop(self._hwnd)
#     def SetAsForegroundWindow(self):
#         self.shell.SendKeys('%')
#         win32gui.SetForegroundWindow(self._hwnd)
#     def Maximize(self):
#         win32gui.ShowWindow(self._hwnd, win32con.SW_MAXIMIZE)
#     def setActWin(self):
#         win32gui.SetActiveWindow(self._hwnd)
#     def _window_enum_callback(self, hwnd, wildcard):
#         '''Pass to win32gui.EnumWindows() to check all the opened windows'''
#         print(f'hwnd={hwnd} \t window text={win32gui.GetWindowText(hwnd)} \t class={win32gui.GetClassName(hwnd)} \t id={win32gui.GetDlgCtrlID(hwnd)}')
#         if re.match(wildcard, str(win32gui.GetWindowText(hwnd))) is not None:
#             self._hwnd = hwnd
#     def find_window_wildcard(self, wildcard, max_attempts=3):
#         self._hwnd = None
#         attempts = 0
#         while self._hwnd is None and attempts <= max_attempts:
#             sleep(2)
#             win32gui.EnumWindows(self._window_enum_callback, wildcard)
#         if self._hwnd is None: raise NotFoundErr
#     def kill_task_manager(self):
#         wildcard = 'Gestionnaire des t.+ches de Windows'
#         self.find_window_wildcard(wildcard)
#         if self._hwnd:
#             win32gui.PostMessage(self._hwnd, win32con.WM_CLOSE, 0, 0)
#             sleep(0.5)
#     def _window_children_enum_callback(self, hwnd, wildcard):
#         print(f'hwnd={hwnd} \t window text={win32gui.GetWindowText(hwnd)} \t class={win32gui.GetClassName(hwnd)} \t id={win32gui.GetDlgCtrlID(hwnd)}')
#         if win32gui.GetClassName(hwnd) == 'ToolbarWindow32' and 'Address:' in win32gui.GetWindowText(hwnd):
#             self._folder_control_id = hwnd
#     def enum_children(self):
#         win32gui.EnumChildWindows(self._hwnd, self._window_children_enum_callback, '')

#     def set_folder(self, folder):
#         current_folder = win32gui.GetWindowText(self._folder_control_id).replace('Address: ','')
#         print(current_folder)
#         if current_folder != folder:
#             self.SetAsForegroundWindow()
#             sleep(1)
#             self.shell.SendKeys(folder)
#             self.shell.SendKeys('{ENTER}')
#             initial_time = time()
#             while current_folder != folder:
#                 current_folder = win32gui.GetWindowText(self._folder_control_id).replace('Address: ','')
#                 sleep(0.2)
#                 if time()-initial_time >= 10:
#                     raise TimeoutError('Timed out waiting for folder')

#     def set_files(self, files):
#         self.SetAsForegroundWindow()
#         sleep(3)
#         self.shell.SendKeys('p')    # placeholder, otherwise it misses the first character of the actual filename
#         self.shell.SendKeys(files)
#         self.shell.SendKeys('{ENTER}')

# def upload_through_windows_dialog(folder, files):
#     cW = cWindow()
#     # cW.kill_task_manager()
#     initial_time = time()
#     cW.find_window_wildcard('Open')
#     cW.BringToTop()
#     cW.SetAsForegroundWindow()
#     cW.enum_children()
#     cW.set_folder(folder)
#     cW.set_files(files)

import logging
from time import time, sleep
import os

from pywinauto.application import Application
from pywinauto.findwindows import ElementNotFoundError
from pywinauto.base_wrapper import ElementNotEnabled


def upload_through_file_dialog(folder, files):
    logging.debug(f'Uploading through file dialog \n {folder=} \n {files=}')

    initial_time = time()
    while True:
        try:
            logging.debug('Locating file dialog app')
            app = Application(backend='win32').connect(title_re="Open", found_index=0)
            break
        except ElementNotFoundError:
            logging.debug('Could not locate app')
            sleep(0.5)
            if time() - initial_time >= 20:
                logging.exception('File dialog app did not appear')
                raise TimeoutError('File dialog app did not appear')

    logging.debug('Locating file dialog')
    file_dlg = app.window(title_re="Open", found_index=0)
    address_bar = file_dlg.window(title_re="Address:*", class_name='ToolbarWindow32', top_level_only=False, found_index=0)
    open_button = file_dlg['&Open']
    current_folder = address_bar.window_text()

    if current_folder != folder:
        logging.debug(f'Setting folder: {folder}')
        file_dlg.FileNameEdit.set_edit_text(folder)
        open_button.click()

        initial_time = time()
        while True:
            logging.debug('Waiting for folder')
            if 'Downloads' in folder and 'Downloads' in current_folder:
                break
            elif folder in current_folder:
                break

            if time() - initial_time >= 10:
                logging.exception('Folder did not load in dialog')
                raise TimeoutError('Folder did not load in dialog')
            
            sleep(0.5)
            current_folder = address_bar.window_text()

    logging.debug(f'Set folder {folder}')
    initial_time = time()
    while True:
        try:
            logging.debug(f'Setting file(s): {files}')
            file_dlg.FileNameEdit.set_edit_text(files)
            open_button.click()
            break
        except ElementNotEnabled:
            logging.debug('Filename edit not yet enabled')
            sleep(0.5)
            if time() - initial_time >= 20:
                logging.exception('Filename edit did not become enabled')
                raise TimeoutError('Filename edit did not become enabled')

def get_file_dialog(title_re, wait_secs=20, ignore_error=False):
    logging.debug(f'Locating app with title regex "{title_re}"')
    initial_time = time()
    while True:
        try:
            logging.debug('Locating file dialog app')
            app = Application(backend='win32').connect(title_re=title_re, found_index=0)
            logging.debug('Locating file dialog')
            file_dlg = app.window(title_re=title_re, found_index=0)
            return file_dlg
        except ElementNotFoundError:
            logging.debug('Could not locate app')
            sleep(0.5)
            if time() - initial_time >= wait_secs:
                if ignore_error:
                    logging.debug(f'File dialog app did not appear but ignored due to {ignore_error=}')    
                    return None          
                else:
                    logging.exception('File dialog app did not appear')
                    raise TimeoutError('File dialog app did not appear')

from pywinauto.timings import TimeoutError as PyWinAutoTimeoutError
from pywinauto.base_wrapper import ElementNotEnabled

def save_in_default_filepath():
    logging.debug('Saving in whichever folder and filename the file dialog opens with')

    file_dlg = get_file_dialog('Save As')

    address_bar = file_dlg.window(title_re='Address:*', class_name='ToolbarWindow32', top_level_only=False, found_index=0)
    current_folder = address_bar.window_text().replace('Address: ','')
    current_filename = file_dlg['Combobox'].element_info.name

    if current_folder == 'Downloads':
        current_folder = f'C:\\Users\\{os.getlogin()}\\Downloads'

    open_button = file_dlg['&Save']

    initial_time = time()
    while True:
        logging.debug('Clicking on Save button until file dialog disappears')
        open_button.click()
        try:
            file_dlg.wait_not('visible', timeout=3)
            logging.debug(f'File dialog closed with folder "{current_folder}" and file "{current_filename}"')    
            return current_folder, current_filename
        except PyWinAutoTimeoutError:
            logging.debug('File dialog still up, retrying...')
            sleep(1)

        if time() - initial_time >= 10:
            e = TimeoutError('Timed out attempting to close file dialog by clicking Save')
            logging.exception(e)
            raise e



def upload(folder, files, timeout=5):
    logging.debug(f'Uploading via file dialog: \n {folder} \n {files}')

    file_dlg = get_file_dialog('Open')
    address_bar = file_dlg.window(title_re='Address:*', class_name='ToolbarWindow32', top_level_only=False, found_index=0)
    file_edit = file_dlg['Combobox']

    initial_time = time()
    while True:
        current_folder = address_bar.window_text().replace('Address: ','')
        logging.debug(f'File dialog displaying folder: {current_folder}')

        if current_folder == folder:
            logging.debug(f'Folder set: {folder}')
            break
        try:
            logging.debug(f'Setting folder: {folder}')
            file_edit.wait('visible', timeout=timeout)
            file_edit.set_focus()
            file_edit.type_keys(folder, with_spaces=True)
            file_edit.type_keys('{ENTER}')
        except PyWinAutoTimeoutError:
            logging.debug('Timed out waiting for file edit to become visible / enabled, retrying...')
        except ElementNotEnabled:
            logging.debug('File edit not enabled yet, retrying...')

        if time() - initial_time >= 30:
            e = TimeoutError('Timed out attempting to set folder in file dialog')
            logging.exception(e)
            raise e
        sleep(3)

    logging.debug(f'Preparing text from files: {files}')
    files_quoted = [f'"{file}"' for file in files]
    files_quoted = ' '.join(files_quoted)
    
    initial_time = time()
    while True:
        try:
            logging.debug(f'Inputting files: {files_quoted}')
            file_edit.wait('visible', timeout=timeout)
            file_edit.set_focus()
            file_edit.type_keys(files_quoted, with_spaces=True)
            file_edit.type_keys('{ENTER}')
            try:
                file_dlg.wait_not('visible', timeout=3)
                logging.debug(f'File dialog closed')    
                return True
            except PyWinAutoTimeoutError:
                logging.debug('File dialog still up, retrying...')
                sleep(1)
        except PyWinAutoTimeoutError:
            logging.debug('Timed out waiting for file edit to become visible / enabled, retrying...')
        except ElementNotEnabled:
            logging.debug('File edit not enabled yet, retrying...')

        if time() - initial_time >= 10:
            e = TimeoutError('Timed out waiting for file dialog to close')
            logging.exception(e)
            raise e
        sleep(1)
